@@ -18,8 +18,12 @@ def rename_kwargs(func_name, kwargs, aliases):  # noqa
     for alias, new in aliases.items():
         if alias in kwargs:
             if new in kwargs:
-                raise TypeError("{} received both {} and {}".format(func_name, alias, new))
-            warnings.warn("{} is deprecated; use {}".format(alias, new), DeprecationWarning, 3)
+                raise TypeError(
+                    "{} received both {} and {}".format(func_name, alias, new)
+                )
+            warnings.warn(
+                "{} is deprecated; use {}".format(alias, new), DeprecationWarning, 3
+            )
 
             if alias == "device":
                 if "cuda" in kwargs[alias]:
